package DX::ShellSession;

use Tcl;
use Scalar::Util qw(weaken);
use DX::Expander;
use DX::Class;

has shell_state => (is => 'rwp', required => 1, isa => ShellState);

has expander => (
  is => 'lazy', builder => sub { DX::Expander->new(tcl => $_[0]->tcl) },
  handles => [ qw(expand_args) ],
);

has tcl => (is => 'lazy', builder => sub {
  my ($self) = @_;
  weaken $self;
  my $tcl = Tcl->new;
  $tcl->CreateCommand('?' => sub {
    $self->apply_to_state([ 'new_query_state' ], [ mode => 'query' ])
  });
  $tcl->CreateCommand('?+' => sub {
    $self->apply_to_state([ mode => 'query' ])
  });
  $tcl->CreateCommand('.' => sub {
    $self->apply_to_state([ mode => 'shell' ])
  });
  foreach my $pred (
    keys %{$self->shell_state->template_query_state->predicates}
  ) {
    $tcl->CreateCommand($pred => sub {
      my (undef, undef, undef, @args) = @_; # ($data, $interp, $pred)
      $self->apply_predicate($pred => @args);
    });
  }
  return $tcl;
});

sub apply_to_state {
  my ($self, @to_apply) = @_;
  my $state = $self->shell_state;
  our @Result;
  foreach my $to_apply (@to_apply) {
    my ($change, @args) = @$to_apply;
    ($state, my @this_result) = $state->${\"with_${change}"}(@args);
    push @Result, @this_result;
  }
  $self->_set_shell_state($state);
  return;
}

sub is_complete_command_string {
  my ($self, $string) = @_;
  return !!$self->tcl->icall(info => complete => $string);
}

sub eval_command_string {
  my ($self, $string) = @_;
  local our @Result;
  try {
    $self->tcl->Eval($string);
  } catch {
    push @Result, [ output => $_ ];
  };
  return @Result;
}

sub apply_predicate {
  my ($self, $pred, @args) = @_;
  die "Can't call predicate ${pred} outside a query\n"
    unless $self->shell_state->mode eq 'query';
use Data::Dumper::Concise; die Dumper [ $pred, $self->expand_args(@args) ];
}

1;
